#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <regex.h>
#include <signal.h>
#include "udmsearch.h"
#include "mysql.h"
#include "indexer.h"
#include "parseurl.h"
#include "openhost.h"
#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#ifdef CHARSET
#include "charset.h"
#endif
#ifndef CONFIGDIR
#define CONFIGDIR "."
#endif

#define ER_DUP_ENTRY		1062
#define CR_SERVER_LOST		2013
#define CR_SERVER_GONE_ERROR	2006
#define ER_SERVER_SHUTDOWN	1053

#define free_it(x)		{if(x){free(x);x=NULL;}}
#define STRNCMP(x,y)		strncmp(x,y,strlen(y))
#define STRNCASECMP(x,y)	strncasecmp(x,y,strlen(y))

#define E_KOI8 0
#define E_WIN  1
#define HTML_UNK 0
#define HTML_TAG 1
#define HTML_TXT 2
#define HTML_COM 3
#define SEPARATOR " \t\r\n\\\" []^_!#$%&'()*+,-./:;<=>?@^_`{|}~"

MYSQL mysql;
char *MySQLHost       = MYSQL_HOST;
char *MySQLDB         = MYSQL_DB;
char *MySQLUser       = MYSQL_USER;
char *MySQLPass       = MYSQL_PASS;
char qbuf[STRSIZ*4];

static int alarm_abort;
static void sig_alarm(int i){ 
	alarm_abort = 1;
}

int dis_rules_count=0;
regex_t dis_rules[MAXFILTER];

int free_disallow(void){
int i;
	for(i=0; i < dis_rules_count; i++)
		regfree(&(dis_rules[i]));
	return(0);
}

int add_disallow(char * rule){
#define ERRSTRSIZE 100
char errstr[ERRSTRSIZE];
int err;
	if(dis_rules_count<MAXFILTER){
		err=regcomp(&(dis_rules[dis_rules_count]),rule,REG_EXTENDED|REG_ICASE);
		if(err){
			fprintf(stdout, "\nConfig file error: wrong 'disallow' regex: %s\n", rule);
			regerror(err, &(dis_rules[dis_rules_count]), errstr, ERRSTRSIZE);
			fprintf(stdout, "Error is: %s\n", errstr);
			regfree(&(dis_rules[dis_rules_count]));
			return(1);
		}
		dis_rules_count++;
	}else{
		fprintf(stdout,"\nToo many 'disallow' commands in config file\nTry to increment MAXFILTER in indexer.h and recompile\n");
		return(1);
	}
	return(0);
}

int is_denied(char *url){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
	for(i=0;i<dis_rules_count;i++){
		err=regexec(&(dis_rules[i]),url,NS,subs,0);
		if(!err)return(1);
	}
	return(0);
}

int head_rules_count=0;
regex_t head_rules[MAXFILTER];

int free_head_rules(void){
int i;
	for(i=0; i < head_rules_count; i++)
		regfree(&(head_rules[i]));
	return(0);
}

int add_head_rule(char * rule){
#define ERRSTRSIZE 100
char errstr[ERRSTRSIZE];
int err;
	if(head_rules_count<MAXFILTER){
		err=regcomp(&(head_rules[head_rules_count]),rule,REG_EXTENDED|REG_ICASE);
		if(err){
			fprintf(stdout, "\nConfig file error: wrong 'CheckOnly' regexx: %s\n", rule);
			regerror(err, &(head_rules[head_rules_count]), errstr, ERRSTRSIZE);
			fprintf(stdout, "Error is: %s\n", errstr);
			regfree(&(head_rules[head_rules_count]));
			return(1);
		}
		head_rules_count++;
	}else{
		fprintf(stdout,"\nToo many 'CheckOnly' commands in config file\nTry to increment MAXFILTER in indexer.h and recompile\n");
		return(1);
	}
	return(0);
}

int is_head_only(char *url){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
	for(i=0;i<head_rules_count;i++){
		err=regexec(&(head_rules[i]),url,NS,subs,0);
		if(!err)return(1);
	}
	return(0);
}

int wcur,wsize;
typedef struct word_struct {
	int count;
	char *word;
} WORD;
WORD *Word=NULL;



int nservers=0;
typedef struct server_struct {
	char *url;
	char *charset;
	char *proxy;
	char *basic_auth;
	int  proxy_port;
	int  period;
	int  hint;
	int  net_errors;
	int  max_net_errors;
	int  outside;
	int  maxhops;
	int  gindex;
	int  gfollow;
	int  deletebad;
	int  userobots;
	int  bodyweight;
	int  titleweight;
	int  urlweight;
	int  descweight;
	int  keywordweight;
} SERVER;
SERVER Server[MAXSERVER];

int cmpserver(const void *s1,const void *s2)
{
	return(strlen(((SERVER*)s2)->url)-strlen(((SERVER*)s1)->url));
}


int nrobots=0;
typedef struct robots_struct {
	char *hostinfo;
	char *path;
} ROBOTS;
ROBOTS Robots[MAXROBOTS];

int in_stop_list=0;
char *stoplist[MAXSTOPWORD];


void abort_indexer()
{
	fprintf(stdout,"Indexer: Error #%d (%s)\n",mysql_errno(&mysql),mysql_error(&mysql));
	exit(-1);
}

int safe_mysql_query(MYSQL *mysql,char *query)
{
int res,i;
	for(i=0;i<2;i++){
		if(res=mysql_query(mysql,query)){
			if((mysql_errno(mysql)==CR_SERVER_LOST)||
			(mysql_errno(mysql)==CR_SERVER_GONE_ERROR)||
			(mysql_errno(mysql)==ER_SERVER_SHUTDOWN)){
				fprintf(stdout,"Error in mysql_query: %d (%s)\n",mysql_errno(mysql),mysql_error(mysql));
				fprintf(stdout,"Waiting during 5 seconds for server to restart...");
				sleep(5);
				fprintf(stdout,"Trying to reconnect\n");
			}else
				return(res);
		}else
			return(res);
	}
	return(res);
}

int is_stopword(char *word)
{
int i;
	for(i=0;i<in_stop_list;i++){
		if(!strcmp(word,stoplist[i]))
			return(1);
	}
	return(0);
}

int addrobots(char *hostinfo,char *path){
	if(nrobots>=MAXROBOTS)return(1);
	Robots[nrobots].hostinfo=strdup(hostinfo);
	Robots[nrobots].path=strdup(path);
	nrobots++;
	return(0);
}

int freerobots(){
	while(nrobots>0){
		free(Robots[nrobots].hostinfo);
		free(Robots[nrobots].path);
		nrobots--;
	}
	return(0);
}

int addserver(char *url,char *charset,char *proxy,
	int period,int hint,char *basic_auth,
	int outside,int maxhops,
	int gindex,int deletebad,int userobots,int gfollow,
	int  bodyweight,int  titleweight,int  urlweight,
	int  descweight,int  keywordweight,int max_net_errors)
{
char *s;
	if(nservers>=MAXSERVER)return(1);
	Server[nservers].url=strdup(url);
	Server[nservers].charset=charset?strdup(charset):NULL;
	Server[nservers].basic_auth=basic_auth?strdup(basic_auth):NULL;
	if(proxy){
		Server[nservers].proxy=strdup(proxy);
		s=strchr(Server[nservers].proxy,':');
		if(s){
			*s=0;
			Server[nservers].proxy_port=atoi(s+1)?atoi(s+1):DEFAULT_PROXY_PORT;
		}else{
			Server[nservers].proxy_port=DEFAULT_PROXY_PORT;
		}
	}else{
		Server[nservers].proxy=NULL;
		Server[nservers].proxy_port=0;
	}
	Server[nservers].period=period;
	Server[nservers].hint=hint;
	Server[nservers].net_errors=0;
	Server[nservers].outside=outside;
	Server[nservers].maxhops=maxhops;
	Server[nservers].gindex=gindex;
	Server[nservers].gfollow=gfollow;
	Server[nservers].deletebad=deletebad;
	Server[nservers].userobots=userobots;
	Server[nservers].bodyweight=bodyweight;
	Server[nservers].titleweight=titleweight;
	Server[nservers].urlweight=urlweight;
	Server[nservers].descweight=descweight;
	Server[nservers].keywordweight=keywordweight;
	Server[nservers].max_net_errors=max_net_errors;
	nservers++;
	return(0);
}

int findserver(char *url){
int i;
char *r;
	r=strstr(url,"/robots.txt");
	if((r)&&(!strcmp(r,"/robots.txt"))){
		*r=0;
		for(i=0;i<nservers;i++){
			if(!STRNCMP(Server[i].url,url)){
				*r='/';
				return(i);
			}
		}
		*r='/';
	}else{
		for(i=0;i<nservers;i++)
			if(!STRNCMP(url,Server[i].url))
				return(i);
	}
	return(-1);
}

int findrobots(char *url){
int i;
	for(i=0;i<nrobots;i++)
		if(strstr(url,Robots[i].hostinfo)&&strstr(url,Robots[i].path))
			return(i);
	return(-1);
}

int addurl(char *url,int referrer,int hops)
{
int i;
char *e_url,*s;
	e_url=(char*)malloc(strlen(url)*2);
	mysql_escape_string(e_url,url,strlen(url));
	sprintf(qbuf,"INSERT INTO url (url,referrer,hops) VALUES ('%s',%d,%d)",e_url,referrer,hops);
	free(e_url);
	if(safe_mysql_query(&mysql,qbuf)){
		if(mysql_errno(&mysql)!=ER_DUP_ENTRY){
			abort_indexer();
		}
	}
	return(0);
}

int deletewordfromurl(int url_id)
{
	sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d",url_id);
	if(safe_mysql_query(&mysql,qbuf)) abort_indexer();
	return(0);
}

int DeleteRobotsFromHost(char *hostinfo)
{
	sprintf(qbuf,"DELETE FROM robots WHERE hostinfo='%s'",hostinfo);
	if(safe_mysql_query(&mysql,qbuf)) abort_indexer();
	return(0);
}

int deleteurl(int url_id)
{
	sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d",url_id);
	if(safe_mysql_query(&mysql,qbuf)) abort_indexer();
	sprintf(qbuf,"DELETE FROM url WHERE rec_id=%d",url_id);
	if(safe_mysql_query(&mysql,qbuf)) abort_indexer();
	return(0);
}

int addword(char *word,int where,int checkstop)
{
int i;
char *e,*s;

	if(!strcmp(word,"")) return(0);
	s=e=strdup(word);
	while(*s){
#ifdef CHARSET
		*s=rus_tolower(*s);
#else
		*s=tolower(*s);
#endif
		s++;
	}
	if(checkstop){
		if(is_stopword(e)){
			free(e);
			return(0);
		}
	}
	for(i=0;i<wcur;i++){
		if((!strcmp(Word[i].word,e))&&(Word[i].count<256)){
			Word[i].count|=where;
			free(e);
			return(0);
		}
	}
	if(wcur>=wsize){
		if(wsize){
			wsize+=WSIZE;
			Word=(WORD *)realloc(Word,wsize*sizeof(WORD));
		}else{
			wsize=WSIZE;
			Word=(WORD *)malloc(wsize*sizeof(WORD));
		}
	}
	Word[wcur].word=strdup(e);
	Word[wcur].count=where;
	wcur++;free(e);
	return(0);
}

int LoadRobotsRules(){
MYSQL_RES *res;
MYSQL_ROW row;
	freerobots();
	if(safe_mysql_query(&mysql,"SELECT hostinfo,path FROM robots"))
		abort_indexer();
	if(res=mysql_store_result(&mysql)){
		while(row=mysql_fetch_row(res))
			addrobots(row[0],row[1]);
		mysql_free_result(res);
	}
	return(0);
}

int storewords(int url_id)
{
int i,j,old,new;
MYSQL_RES *res;
MYSQL_ROW row;
int were,now,added,deleted,updated;

	were=now=added=deleted=updated=0;
	now=wcur;
	sprintf(qbuf,"SELECT word,intag FROM dict WHERE url_id=%d",url_id);
	if(safe_mysql_query(&mysql,qbuf))abort_indexer();
	if(res=mysql_store_result(&mysql)){
		while(row=mysql_fetch_row(res)){
			if(!row[0])continue;
			if(!row[1])continue;
			addword(row[0],atoi(row[1])*256,0);
			were++;
		}
		mysql_free_result(res);
	}
	sprintf(qbuf,"LOCK TABLES dict WRITE");
	if(safe_mysql_query(&mysql,qbuf))abort_indexer();

	j=0;
	for(i=0;i<wcur;i++){
		if(j>1024){
			j=0;
			sprintf(qbuf,"UNLOCK TABLES");
			if(safe_mysql_query(&mysql,qbuf))abort_indexer();
			sprintf(qbuf,"LOCK TABLES dict WRITE");
			if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		}
		old=(Word[i].count/256)&255;
		new=Word[i].count&255;
		if((!new)&&(old)){
			sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d AND word='%s'",url_id,Word[i].word);
			if(safe_mysql_query(&mysql,qbuf))abort_indexer();
			j++; deleted++;
			continue;
		}
		if((new)&&(!old)){
			sprintf(qbuf,"INSERT INTO dict (url_id,word,intag) VALUES(%d,'%s',%d)",url_id,Word[i].word,Word[i].count);
			if(safe_mysql_query(&mysql,qbuf))abort_indexer();
			j++; added++;
			continue;
		}
		if((new)&&(old)&&(new!=old)){
			sprintf(qbuf,"UPDATE dict SET intag=%d WHERE word='%s' AND url_id='%d'",new,Word[i].word,url_id);
			if(safe_mysql_query(&mysql,qbuf))abort_indexer();
			j++; updated++;
			continue;
		}
	}
	sprintf(qbuf,"UNLOCK TABLES");
	if(safe_mysql_query(&mysql,qbuf))abort_indexer();
	return(0);
}

int freewords() {
	int i;
	for(i=0;i<wcur;i++) free(Word[i].word);
	wcur=0;
	return(0);
}

int is_connected=0;
int connect_db()
{
#if MYSQL_VERSION_ID >= 32200
	if (!(mysql_real_connect(&mysql,MySQLHost,MySQLUser,MySQLPass,MySQLDB,0,NULL,0)))abort_indexer();
#else
	if (!(mysql_connect(&mysql,MySQLHost,MySQLUser,MySQLPass)))abort_indexer();
	if (mysql_select_db(&mysql,MySQLDB))abort_indexer();
#endif
	is_connected++;
	return(0);
}


#define BASE64_LEN(len) (4 * (((len) + 2) / 3) +2)
/* BASE64 encoding: 3x8 bits -> 4x6 bits */
static char base64[64] = 
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

void base64_encode (const char *s, char *store, int length)
{
int i;
unsigned char *p = (unsigned char *)store;

	for (i = 0; i < length; i += 3){
		*p++ = base64[s[0] >> 2];
		*p++ = base64[((s[0] & 3) << 4) + (s[1] >> 4)];
		*p++ = base64[((s[1] & 0xf) << 2) + (s[2] >> 6)];
		*p++ = base64[s[2] & 0x3f];
		s += 3;
	}
	/*Pad the result*/
	if (i == length + 1) *(p - 1) = '=';
	else if (i == length + 2) *(p - 1) = *(p - 2) = '=';
	*p = '\0';
}

char *HTTPErrMsg(int code){
	switch(code){
	case 0:   return("Not indexed yet");
	case 200: return("OK");
	case 301: return("Moved Permanently");
	case 302: return("Moved Temporarily");
	case 303: return("See Other");
	case 304: return("Not Modified");
	case 300: return("Multiple Choices");
	case 305: return("Use Proxy (proxy redirect)");
	case 400: return("Bad Request");
	case 401: return("Unathorized");
	case 402: return("Payment Required");
	case 403: return("Forbidden");
	case 404: return("Not found");
	case 405: return("Method Not Allowed");
	case 406: return("Not Acceptable");
	case 407: return("Proxy Authentication Required");
	case 408: return("Request Timeout");
	case 409: return("Conflict");
	case 410: return("Gone");
	case 411: return("Length Required");
	case 412: return("Precondition Failed");
	case 413: return("Request Entity Too Large");
	case 414: return("Request-URI Too Long");
	case 415: return("Unsupported Media Type");
	case 500: return("Internal Server Error");
	case 501: return("Not Implemented");
	case 502: return("Bad Gateway");
	case 505: return("HTTP Version Not Supported");
	case 503: return("Service Unavailable");
	case 504: return("Gateway Timeout");
	default:  return("Unknown status");
	}
}

int HTTPGet(char *host,int port,char *header,char *dist,int maxsize){
int fd,i,size;
char *s;
void *old_alarm;

	alarm_abort=0;alarm(READ_TIMEOUT);
	old_alarm=signal(SIGALRM,sig_alarm);

	/* Connect to HTTP or PROXY server */
	fd=open_host(host,port);	/* Connect           */
	if(fd<0)return(fd);		/* Could not connect */
	write(fd,header,strlen(header));/* Send HTTP request */
	size=0;s=dist;			/* Retrieve response */
	while(i=read(fd,s,maxsize-1-size)){
		size+=i;s+=i;
		if(alarm_abort)
			break;
	}
	close(fd);			/* Close HTTP connection */
	alarm(0);signal(SIGALRM,old_alarm);
	if(alarm_abort)return(-3);
	return(size);
}

int usage(){
	fprintf(stderr,
"
indexer for UdmSearch Distrib %s,
(C) 1998-1999, UdmSearch Developers Team, Alexander Barkov (bar@izhcom.ru)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Usage:
	indexer [OPTIONS]  [configfile]

	-a		reindex URLs even if not expired yet
	-n number	reindex only number documents and exit
	-e		more expired first

	-t tag		set filter on this Tag
	-u pattern	set filter on Url (MySQL LIKE wildcards)
	-s status	set filter on status

	-C		Clear database and exit
	-S		show Statistics and exit
	-I		show referrers and exit
	-h,-?		show this help page

",VERSION);
	return(0);
}

main(int argc, char **argv) {
int   i,j,state,len;
char  tagstr[STRSIZ]="";
char  urlstr[STRSIZ]="";
char  statusstr[STRSIZ]="";
char  urlin[STRSIZ]="";
char  str[STRSIZ]="";
char  str1[STRSIZ]="";
char text[2*MAXTEXTSIZE]="";
char text_escaped[2*MAXTEXTSIZE]="";
char keywords[MAXKEYWORDSIZE]="";
char keywords_escaped[2*MAXKEYWORDSIZE]="";
char descript[2*MAXDESCSIZE]="";
char descript_escaped[2*MAXDESCSIZE]="";
char title[2*MAXTITLESIZE]="";
char get_url_buf[STRSIZ]="";
char user_agent[256]="";

MYSQL_RES *urlres;
MYSQL_RES *res;
MYSQL_ROW row;
int currow,nrows;
FILE  *config;
URL from,newurl;
TAG tag;
char *buf,*s,*e,*tmp;
int url_id,size,oldsize,oldstatus,status,hops,hint,realsize;
int period=DEFAULT_REINDEX_TIME;
int maxhops=DEFAULT_MAX_HOPS;
int outside=0;

char last_index_time[255];
char cururl[STRSIZ]="";
char *content,*header,*content_type,*location,*href;
int  inbody,intitle,inscript,follow,index;
char *charset=NULL,*proxy=NULL,*basic_auth=NULL;
char *config_name= CONFIGDIR "/indexer.conf";
int gindex=1;
int gfollow=1;
int deletebad=1;
int userobots=1;
int npages=-1;
int npages0=0;
int efirst=0;
int max_net_errors=MAXNETERRORS;
int bodyweight      = 1;
int titleweight     = 2;
int keywordweight   = 2;
int descweight      = 2;
int urlweight       = 0;

extern char *optarg;
extern int optind;
int ch;
int reindexall=0;
int done=0;
int clear=0;
int stat=0;
int integrity=0;
int indexed_total=0;
int total_total=0;
int reload_robots=1;
int myrule=0;
int headonly=0;
#define URL_SELECT_CACHE 50
#define URL_DELETE_CACHE 50

	/* Parse arguments */
	while ((ch = getopt(argc, argv, "CSIahe?t:u:s:n:")) != -1){
		switch (ch) {
		case 'C':
			clear++;
			break;
		case 'S':
			stat++;
			break;
		case 'I':
			integrity++;
			break;
		case 't':
			i=atoi(optarg);
			sprintf(str,"%d",i);
			if(*tagstr){
				strcat(tagstr,",");
			}else{
				strcat(tagstr," AND url.tag IN (");
			}
			strcat(tagstr,str);
			break;
		case 's':
			i=atoi(optarg);
			sprintf(str,"%d",i);
			if(*statusstr){
				strcat(statusstr,",");
			}else{
				strcat(statusstr," AND url.status IN (");
			}
			strcat(statusstr,str);
			break;
		case 'u':
			if(*urlstr){
				strcat(urlstr," OR ");
			}else{
				strcat(urlstr," AND (");
			}
			strcat(urlstr,"url.url LIKE '");
			strcat(urlstr,optarg);
			strcat(urlstr,"'");
			break;
		case 'a':
			reindexall=1;
			break;
		case 'e':
			efirst=1;
			break;
		case 'n':
			npages=atoi(optarg);
			npages0=npages;
			break;
		case '?':
		case 'h':
		default:
			usage();
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;
	if(argc>1){
		usage();
		exit(1);
	}

	if(*tagstr)strcat(tagstr,")");
	if(*statusstr)strcat(statusstr,")");
	if(*urlstr)strcat(urlstr,")");

	/* Load config file */
	if (argc==1)
	    config_name=argv[0];
	fprintf(stdout,"Loading '%s' as config file...",config_name);

	if(!(config=fopen(config_name,"r"))){
		fprintf(stdout,"Can't open config file\n");
		exit(1);
	}
	mysql_init(&mysql);

	i=0;
	while(fgets(str,sizeof(str),config)){
		i++; strcpy(str1,"");
		if(!strcmp(str,"\n"))
			continue;
		if(str[0]=='#')
			continue;
		if(!STRNCASECMP(str,"MySQLHost")){
			if(sscanf(str+9,"%s",str1))
				MySQLHost=strdup(str1);
		}else
		if(!STRNCASECMP(str,"MySQLDB")){
			if(sscanf(str+7,"%s",str1)) 
				MySQLDB=strdup(str1);
		}else
		if(!STRNCASECMP(str,"MySQLUser")){
			if(sscanf(str+9,"%s",str1)&&strcmp(str1,""))
				MySQLUser=strdup(str1);
		}else
		if(!STRNCASECMP(str,"MySQLPass")){
			if(sscanf(str+9,"%s",str1)&&strcmp(str1,""))
				MySQLPass=strdup(str1);
			else
				MySQLPass=strdup("");
		}else
		if(!STRNCASECMP(str,"disallow")){
			s=strtok(str," \t\r\n");
			while(s=strtok(NULL," \t\r\n")){
				if(add_disallow(s))
					exit(1);
			}
		}else
		if(!STRNCASECMP(str,"CheckOnly")){
			s=strtok(str," \t\r\n");
			while(s=strtok(NULL," \t\r\n")){
				if(add_head_rule(s))
					exit(1);
			}
		}else
		if(!STRNCASECMP(str,"Tag")){
			if(sscanf(str+3,"%d",&hint)!=1)
				hint=0;
		}else
		if(!STRNCASECMP(str,"MaxNetErrors")){
			if(sscanf(str+12,"%d",&max_net_errors)!=1)
				max_net_errors=MAXNETERRORS;
		}else
		if(!STRNCASECMP(str,"Period")){
			if(sscanf(str+7,"%d",&period)!=1)
				period=DEFAULT_REINDEX_TIME;
		}else
#ifdef CHARSET
		if(!STRNCASECMP(str,"CharSet")){
			free_it(charset);
			if(sscanf(str+7,"%s",str1))
				charset=strdup(str1);
		}else
#endif
		if(!STRNCASECMP(str,"Proxy")){
			free_it(proxy);
			if(sscanf(str+6,"%s",str1)&&strcmp(str1,""))
				proxy=strdup(str1);
		}else
		if(!STRNCASECMP(str,"AuthBasic")){
			free_it(basic_auth);
			if(sscanf(str+9,"%s",str1)){
				basic_auth=(char*)malloc(BASE64_LEN(strlen(str1)));
				base64_encode(str1,basic_auth,strlen(str1));
			}
		}else
		if(!STRNCASECMP(str,"Server")){
			if(sscanf(str+6,"%s",str1)==1){
				if(!is_connected) connect_db();
				if(!parseurl(&from,str1)){
					addserver(str1,charset,proxy,period,
					hint,basic_auth,outside,maxhops,
					gindex,deletebad,userobots,gfollow,
					bodyweight,titleweight,urlweight,
					descweight,keywordweight,max_net_errors);
					addurl(str1,0,0);
					if(!strcmp(from.schema,"http")){
						sprintf(str,"%s://%s/%s",
							from.schema,
							from.hostinfo,
							"robots.txt");
						addurl(str,0,0);
					}
				}
			}else{
				fprintf(stdout,"Error in config file line %d:\n%s\n",i,str);
				exit(1);
			}
		}else
		if(!STRNCASECMP(str,"FollowOutside")){
			if(sscanf(str+13,"%s",&str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					outside=1;
				else
					outside=0;
			}else
				outside=0;
		}else
		if(!STRNCASECMP(str,"Index")){
			if(sscanf(str+5,"%s",&str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					gindex=1;
				else
					gindex=0;
			}else
				gindex=1;
		}else
		if(!STRNCASECMP(str,"Follow ")){
			if(sscanf(str+7,"%s",&str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					gfollow=1;
				else
					gfollow=0;
			}else
				gfollow=1;
		}else
		if(!STRNCASECMP(str,"Robots")){
			if(sscanf(str+6,"%s",&str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					userobots=1;
				else
					userobots=0;
			}else
				userobots=1;
		}else
		if(!STRNCASECMP(str,"DeleteBad")){
			if(sscanf(str+9,"%s",&str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					deletebad=1;
				else
					deletebad=0;
			}else
				deletebad=1;
		}else
		if(!STRNCASECMP(str,"MaxHops")){
			if(sscanf(str+7,"%d",&maxhops)!=1)
				maxhops=0;
		}else
		if(!STRNCASECMP(str,"BodyWeight")){
			if(sscanf(str+10,"%d",&bodyweight)!=1)
				bodyweight=1;
		}else
		if(!STRNCASECMP(str,"TitleWeight")){
			if(sscanf(str+11,"%d",&titleweight)!=1)
				titleweight=2;
		}else
		if(!STRNCASECMP(str,"DescWeight")){
			if(sscanf(str+10,"%d",&descweight)!=1)
				descweight=2;
		}else
		if(!STRNCASECMP(str,"UrlWeight")){
			if(sscanf(str+9,"%d",&urlweight)!=1)
				urlweight=0;
		}else
		if(!STRNCASECMP(str,"KeywordWeight")){
			if(sscanf(str+13,"%d",&keywordweight)!=1)
				keywordweight=2;
		}else
		if(!STRNCASECMP(str,"NoIndex")){
			gindex=0;
		}else
		if(!STRNCASECMP(str,"NoFollow")){
			gfollow=0;
		}else{
			fprintf(stdout,"Error in config file line %d:\n%s\n",i,str);
			exit(1);
		}
	}
	fclose(config);
	free_it(charset);
	free_it(proxy);
	free_it(basic_auth);
	fprintf(stdout,"Ok\n");
	if(!is_connected) connect_db();

	if(clear){
		printf("You are going to delete from database\n");
		printf("Are you shure?(YES/no)\n");
		if(fgets(str,sizeof(str),stdin)){
			if(strncmp(str,"YES",3))
				exit(0);
		}else
			exit(0);

		if(!tagstr[0]&&!urlstr[0]&&!statusstr[0]){
			printf("Deleting all records\n");
			sprintf(qbuf,"DELETE FROM dict");
			if(safe_mysql_query(&mysql,qbuf))abort_indexer();
			sprintf(qbuf,"DELETE FROM url");
			if(safe_mysql_query(&mysql,qbuf))abort_indexer();
			printf("Done\n");
		}else{
			printf("Deleting\n");j=0;
			while(1){
				sprintf(qbuf,"SELECT rec_id FROM url WHERE 1=1 %s %s %s LIMIT %d",tagstr,urlstr,statusstr,URL_DELETE_CACHE);
				if(safe_mysql_query(&mysql,qbuf))abort_indexer();
				if(res=mysql_store_result(&mysql)){
					if(!mysql_num_rows(res))
						break;
					urlin[0]=0;
					for(i=0;i<mysql_num_rows(res);i++){
						row=mysql_fetch_row(res);
						if(i)strcat(urlin,",");
						strcat(urlin,row[0]);
					}
					j+=i;
					mysql_free_result(res);
					sprintf(qbuf,"DELETE FROM dict WHERE url_id in (%s)",urlin);
					if(safe_mysql_query(&mysql,qbuf))abort_indexer();
					sprintf(qbuf,"DELETE FROM url WHERE rec_id in (%s)",urlin);
					if(safe_mysql_query(&mysql,qbuf))abort_indexer();
					printf("%d records\r",j);fflush(stdout);
				}
			}
			printf("\nDone\n");
		}
		exit(0);
	}

	
	/* Load stop-list */
	if(safe_mysql_query(&mysql,"SELECT word FROM stopword"))abort_indexer();
	if(res=mysql_store_result(&mysql)){
		in_stop_list=mysql_num_rows(res);
		for(i=0;i<in_stop_list;i++){
			row=mysql_fetch_row(res);
			stoplist[i]=strdup(row[0]);
		}
		mysql_free_result(res);
	}
	if(stat){
		printf("\n          UdmSearch statistics\n\n");
		sprintf(qbuf,"SELECT status,sum(next_index_time<=NOW()),count(*) FROM url WHERE 1=1 %s %s %s GROUP BY status",tagstr,urlstr,statusstr);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		if(res=mysql_store_result(&mysql)){
			indexed_total=total_total=0;
			j=mysql_num_rows(res);

			printf("%10s %10s %10s\n","Status","Expired","Total");
			printf("   -----------------------------\n");
			for(i=0;i<j;i++){
				row=mysql_fetch_row(res);
				printf("%10s %10s %10s %s\n",
					row[0],row[1],row[2],
					HTTPErrMsg(atoi(row[0])));
				indexed_total+=atoi(row[1]);
				total_total+=atoi(row[2]);
			}
			mysql_free_result(res);
			printf("   -----------------------------\n");
			printf("%10s %10d %10d\n\n","Total",indexed_total,total_total);
		}
		exit(0);
	}
	
	if(integrity){
		printf("\n          URLs and referrers \n\n");
		sprintf(qbuf,"SELECT url2.url,url.url,url.status FROM url,url url2 WHERE url.referrer=url2.rec_id %s %s %s ",tagstr,urlstr,statusstr);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		if(res=mysql_store_result(&mysql)){
			while(row=mysql_fetch_row(res)){
				printf("%s -> %s (%s)\n",
					row[0],row[1],row[2]);
			}
			mysql_free_result(res);
		}
		exit(0);
	}

	/* Add one virtual server if we want follow outside */
	if(outside)
		addserver("",charset,proxy,period,hint,
			basic_auth,outside,maxhops,gindex,
			deletebad,userobots,gfollow,
			bodyweight,titleweight,urlweight,
			descweight,keywordweight,max_net_errors);

	/*  Long name should be found first    */
	/*  to allow different options         */
	/*  for server and it's subdirectories */
	qsort((void*)Server,nservers,sizeof(SERVER),cmpserver);

	if(reindexall){ /* option -a given */
		sprintf(qbuf,"UPDATE url SET next_index_time=NOW() WHERE 1=1 %s%s%s",tagstr,urlstr,statusstr);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
	}

	/* Some usefull string constants */
	sprintf(get_url_buf,
"SELECT url,rec_id,size,status,concat(date_format(unix_timestamp(last_index_time)+3600*%d,'%%a, %%e %%b %%Y %%T'),' GMT'),hops FROM url WHERE next_index_time<=NOW() %s %s %s %s LIMIT %d",
	GMT_OFFSET,tagstr,urlstr,statusstr,efirst?" ORDER BY next_index_time ":"",URL_SELECT_CACHE);
	sprintf(user_agent,"%s/%s",INDEXER_USER_AGENT,VERSION);

	/* No url retrieved yet */
	currow=1;nrows=0;urlres=NULL;

	/* Alloc buffer for document */
	if(!buf)buf=(char*)malloc(MAXDOCSIZE);

/***************************************************************/
/****************** Now start indexing ... *********************/
/***************************************************************/

while(1){
	if(!npages){
		printf("%d documents done\n",npages0);
		exit(0);
	}
	if(npages>0)npages--;

	if(reload_robots){
		LoadRobotsRules();
		reload_robots=0;
	}
	if(currow>=nrows){  /* Get next URL_CACHE records */
		if(urlres)mysql_free_result(urlres);
		if(safe_mysql_query(&mysql,"LOCK TABLES url WRITE"))abort_indexer();
		if(safe_mysql_query(&mysql,get_url_buf))abort_indexer();
		if(!(urlres=mysql_store_result(&mysql)))
			done=1;
		else
		if(!(mysql_num_rows(urlres))){
			mysql_free_result(urlres);
			done=1;
		}
		if(done){
			if(safe_mysql_query(&mysql,"UNLOCK TABLES"))abort_indexer();
			fprintf(stdout,"Done\n");
			mysql_close(&mysql);
			return(0);
		}
		urlin[0]=0;
		while(row=mysql_fetch_row(urlres)){
			if(urlin[0])strcat(urlin,",");
			strcat(urlin,row[1]);
		}
		sprintf(qbuf,"UPDATE url SET next_index_time=from_unixtime(unix_timestamp()+3600) WHERE rec_id in (%s)",urlin);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		if(safe_mysql_query(&mysql,"UNLOCK TABLES"))abort_indexer();
		currow=0;
		nrows=mysql_num_rows(urlres);
		mysql_data_seek(urlres,0);
	}

	status=1;inbody=intitle=0;
	content_type=location=href=charset=NULL;
	last_index_time[0]=0;text[0]=0;title[0]=0;cururl[0]=0;
	keywords[0]=0;descript[0]=0;

	if(row=mysql_fetch_row(urlres)){
		strncpy(cururl,row[0]?row[0]:"",sizeof(cururl));
		url_id=atoi(row[1]);
		oldsize=row[2]?atoi(row[2]):0;
		oldstatus=row[3]?atoi(row[3]):1;
		strncpy(last_index_time,row[4]?row[4]:"",sizeof(last_index_time));
		hops=row[5]?atoi(row[5]):0;
		done=0;
	}else{
		printf("Fatal error\n");exit(1);
	}
	currow++;

	fprintf(stdout,"%s ",cururl);fflush(stdout);

#ifdef HAVE_SETPROCTITLE
	setproctitle("%s",cururl);
#endif
	if(parseurl(&from,cururl)){
		deleteurl(url_id);
		fprintf(stdout,"Invalid http or ftp url: %s ... deleted\n",cururl);
		freeurl(&from);
		continue;
	}
	if((i=findserver(cururl))<0){
		deleteurl(url_id);
		if(!strcmp(from.filename,"robots.txt")){
			DeleteRobotsFromHost(from.hostinfo);
			reload_robots=1;
		}
		fprintf(stdout,"No server record for url... deleted.\n");
		freeurl(&from);
		continue;
	}
	period=Server[i].period;
	hint=Server[i].hint;
	outside=Server[i].outside;
	maxhops=Server[i].maxhops;
	gindex=Server[i].gindex;index=gindex;
	gfollow=Server[i].gfollow;follow=gfollow;
	userobots=Server[i].userobots;
	deletebad=Server[i].deletebad;
	bodyweight=Server[i].bodyweight;
	titleweight=Server[i].titleweight;
	urlweight=Server[i].urlweight;
	descweight=Server[i].descweight;
	keywordweight=Server[i].keywordweight;

#ifdef CHARSET
	charset=Server[i].charset;
#endif

	if(hops>maxhops){
		deleteurl(url_id);
		fprintf(stdout,"Too many hops (%d)... deleted.\n",hops);
		freeurl(&from);
		continue;
	}
	if(is_denied(cururl)){
		deleteurl(url_id);
		if(!strcmp(from.filename,"robots.txt")){
			DeleteRobotsFromHost(from.hostinfo);
			reload_robots=1;
		}
		fprintf(stdout,"Url disallowed by configuration rule... deleted.\n");
		freeurl(&from);
		continue;
	}
	if((userobots)&&(findrobots(cururl)>=0)){
		/* if robots.txt disallowed by himself */
		/* we will not delete it */
		if(strcmp(from.filename,"robots.txt")){
			deleteurl(url_id);
			fprintf(stdout,"Url disallowed by robots.txt rule... deleted.\n");
		}
		freeurl(&from);
		continue;
	}
	if(strcmp(from.schema,"http")&&strcmp(from.schema,"ftp")){
		deleteurl(url_id);
		fprintf(stdout,"Unknown url schema: %s\n",cururl);
		freeurl(&from);
		continue;
	}
	if((!strcmp(from.schema,"ftp"))&&(!Server[i].proxy)){
		fprintf(stdout,"ftp is available via proxy only: %s\n",cururl);
		sprintf(qbuf,"UPDATE url SET status=503,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		freeurl(&from);
		continue;
	}
	if(Server[i].net_errors>Server[i].max_net_errors){
		fprintf(stdout,"Too many network errors for this server... will try later\n");
		sprintf(qbuf,"UPDATE url SET status=503,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		freeurl(&from);
		continue;
	}
	/* Compose HTTP header */
	headonly=is_head_only(cururl);
	sprintf(str,"%s %s%s HTTP/1.0\r\nHost: %s\r\n",
		headonly?"HEAD":"GET",
		Server[i].proxy?cururl:from.path,
		Server[i].proxy?"":from.filename,
		from.hostname);
	/* Auth if required */
	if(Server[i].basic_auth){
		strcat(str,"Authorization: Basic ");
		strcat(str,Server[i].basic_auth);
		strcat(str,"\r\n");
	}
	
	/* User agent */
	strcat(str,"User-Agent: ");strcat(str,user_agent);strcat(str,"\r\n");
	
	/* If document already indexed, may be it is not modified */
	if(oldstatus&&last_index_time[0]){
		strcat(str,"If-Modified-Since: ");
		strcat(str,last_index_time);
		strcat(str,"\r\n");
	}
#ifdef CHARSET
	/* We prefer koi8-r encoding */
	strcat(str,"Accept-charset: koi8-r");
	strcat(str,"\r\n");
#endif
	/* Empty line is the end of HTTP header */
	strcat(str,"\r\n");

	size = HTTPGet(
		Server[i].proxy?Server[i].proxy:from.hostname,
		Server[i].proxy?Server[i].proxy_port:(from.port?from.port:80),
		str,buf,MAXDOCSIZE
	);

	if(size==-3){
		fprintf(stdout,"Download timeout\n");
		sprintf(qbuf,"UPDATE url SET status=504,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		Server[i].net_errors++;
		freeurl(&from);
		continue;
	}
	if(size<=0){
		fprintf(stdout,"Could not connect to host %s:%d\n",
			Server[i].proxy?Server[i].proxy:from.hostname,
			Server[i].proxy?Server[i].proxy_port:(from.port?from.port:80));
		sprintf(qbuf,"UPDATE url SET status=503,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		Server[i].net_errors++;
		freeurl(&from);
		continue;
	}

	/* Parse response header */
	buf[size]=0;header=buf;
	if(!(content=strstr(buf,"\r\n\r\n")))content=strstr(buf,"\n\n");
	if(!content){
		sprintf(qbuf,"UPDATE url SET status=503,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		freeurl(&from);
		fprintf(stdout,"No content!\n");
		Server[i].net_errors++;
		continue;
	}

	*content=0; content+=4; size-=(content-buf);
	realsize=size;
	s=strtok(header,"\r\n");
	while(s){
		if(!STRNCASECMP(s,"HTTP/")){
			fprintf(stdout,"%s",s); /* !!! */
			status=atoi(s+8);
		}else
		if(!STRNCASECMP(s,"Content-Type: ")){
			content_type=s+14;
#ifdef CHARSET
		if(e=strstr(content_type,"charset="))
			charset=e+8;
#endif
		}else
		if(!STRNCASECMP(s,"Location: ")){
			location=s+10;
		}else
		if(!STRNCASECMP(s,"Content-Length: ")){
			realsize=atoi(s+16);
		}
		s=strtok(NULL,"\r\n");
	}
	printf("\n");

	/* Now parse HTTP code*/
	switch(status){
	case 1: /* No HTTP code */
		/* we'll retry later */
		sprintf(qbuf,"UPDATE url SET status=%d,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",status,period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		fprintf(stdout,"No http code!\n");
		Server[i].net_errors++;
		freeurl(&from);
		continue;

	case 200: /* OK */
		if(!content_type){
			sprintf(qbuf,"UPDATE url SET status=503,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",period,url_id);
			if(safe_mysql_query(&mysql,qbuf))abort_indexer();
			fprintf(stdout,"No Content-type in '%s'!\n",cururl);
			Server[i].net_errors++;
			freeurl(&from);
			continue;
		}
		break;

	case 301:  /* Moved Permanently */
	case 302:  /* Moved Temporarily */
	case 303:  /* See Other */
		/* We'll try to use Location: xxx instead */
		deletewordfromurl(url_id);
		if(hops<maxhops)
		if((location)&&(!is_denied(cururl))){
			if(!parseurl(&newurl,location)){
				if(findserver(location)>=0)
					addurl(location,url_id,hops+1);
				freeurl(&newurl);
			}else{
				fprintf(stdout,"Error in url: '%s'\n",location);
			}
		}
		sprintf(qbuf,"UPDATE url SET status=%d,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",status,period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		freeurl(&from);
		continue;

	case 304:  /* Not Modified, it's very good */
		sprintf(qbuf,"UPDATE url SET status=%d,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",status,period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		freeurl(&from);
		continue;

	case 300:  /* Multiple Choices              */
	case 305:  /* Use Proxy (proxy redirect)    */
	case 400:  /* Bad Request ??? We tried ...  */
	case 401:  /* Unathorized                   */
	case 402:  /* Payment Required              */
	case 403:  /* Forbidden                     */
	case 404:  /* Not found                     */
	case 405:  /* Method Not Allowed            */
	case 406:  /* Not Acceptable                */
	case 407:  /* Proxy Authentication Required */
	case 408:  /* Request Timeout               */
	case 409:  /* Conflict                      */
	case 410:  /* Gone                          */
	case 411:  /* Length Required               */
	case 412:  /* Precondition Failed           */
	case 413:  /* Request Entity Too Large      */
	case 414:  /* Request-URI Too Long          */
	case 415:  /* Unsupported Media Type        */
	case 500:  /* Internal Server Error         */
	case 501:  /* Not Implemented               */
	case 502:  /* Bad Gateway                   */
	case 505:  /* HTTP Version Not Supported    */
		/* delete it if not robots.txt */
	if(deletebad){
		if(!strcmp(from.filename,"robots.txt")){
			sprintf(qbuf,"UPDATE url SET status=%d,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",status,period,url_id);
			if(safe_mysql_query(&mysql,qbuf))abort_indexer();
			DeleteRobotsFromHost(from.hostinfo);
			reload_robots=1;
		}else{
			deleteurl(url_id);
		}
	}else{
		sprintf(qbuf,"UPDATE url SET status=%d,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",status,period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
	}
		freeurl(&from);
		continue;

	case 503: /* Service Unavailable */
	case 504: /* Gateway Timeout */
		/* We'll retry later, maybe server is down */
		sprintf(qbuf,"UPDATE url SET status=%d,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",status,period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		Server[i].net_errors++;
		freeurl(&from);
		continue;

	default:
		/* retry later */
		fprintf(stdout,"HTTP %d We don't yet know how to handle it!\n",status);
		sprintf(qbuf,"UPDATE url SET status=%d,last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d) WHERE rec_id='%d'",status,period,url_id);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		freeurl(&from);
		continue;
	}


	/* Now we have HTTP 200 and know content type */
	
	
	/* robots.txt */
	if(!STRNCASECMP(content_type,"text/plain")&&
	(!strcmp(from.filename,"robots.txt"))){
		sprintf(qbuf,"DELETE FROM robots WHERE hostinfo='%s'",from.hostinfo);
		if(safe_mysql_query(&mysql,qbuf))abort_indexer();
		myrule=0;
		s=strtok(content,"\r\n");
		while(s){
			if(*s=='#'){
			}else
			if(!(STRNCASECMP(s,"User-Agent"))){
				if(strstr(s,INDEXER_USER_AGENT))myrule=1;
				else
				if(strstr(s,"*"))myrule=1;
				else
					myrule=0;
			}else
			if((!(STRNCASECMP(s,"Disallow")))&&(myrule)){
				e=s+9;
				sscanf(e,"%s",str);
				if(str[0]!='#'){
					sprintf(qbuf,"INSERT INTO robots (hostinfo,path) VALUES ('%s','%s')",
						from.hostinfo,str);
					if(safe_mysql_query(&mysql,qbuf))abort_indexer();
				}
			}
			s=strtok(NULL,"\n\r");
		}
		reload_robots=1;
	}else

	/* plain TEXT or the same */
	if(!STRNCASECMP(content_type,"text/plain")||
	!STRNCASECMP(content_type,"text/tab-separated-values")||
	!STRNCASECMP(content_type,"text/css"))
	{
		if(index){
#ifdef CHARSET
			if(charset&&(strstr(charset,"1251")))win2koi(content);
#endif
			strncpy(text,content,sizeof(text));
			s=strtok(content,SEPARATOR);
			while(s){
				if(bodyweight)addword(s,bodyweight,1);
				s=strtok(NULL,SEPARATOR);
			}
		}
	}else

	/* HTML text */
	if(!STRNCASECMP(content_type,"text/html")){
#ifdef CHARSET
		if(charset&&(strstr(charset,"1251")))win2koi(content);
#endif
		state=HTML_UNK; s=content;
		while(*s){
		switch(state){
		case HTML_UNK: /* yet unknown */
			if(!strncmp(s,"<!--",4)){
				state=HTML_COM;
				break;
			}
			if(*s=='<'){
				state=HTML_TAG;
				break;
			}
			state=HTML_TXT;
			break;

		case HTML_COM: /* comment */
			e=strstr(s,"-->");
			if(e)len=e-s+3;
			else len=strlen(s);
			s=s+len; state=HTML_UNK;
			break;

		case HTML_TAG: /* tag */
			e=strchr(s,'>');
			if(e)len=e-s+1;
			else len=strlen(s);
			tmp=(char*)malloc(len+1);
			strncpy(tmp,s,len);tmp[len]=0;
			parsetag(&tag,tmp);

			if(!strcasecmp(tag.name,"body"))inbody=1;
			else
			if(!strcasecmp(tag.name,"/body"))inbody=0;
			else
			if(!strcasecmp(tag.name,"script"))inscript=1;
			else
			if(!strcasecmp(tag.name,"/script"))inscript=0;
			else
			if(!strcasecmp(tag.name,"title"))intitle=1;
			else
			if(!strcasecmp(tag.name,"/title"))intitle=0;
			else
			if((!strcasecmp(tag.name,"a"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"area"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"frame"))&&(tag.src))
				href=strdup(tag.src);
			else
			if((!strcasecmp(tag.name,"meta"))&&(tag.equiv)&&(tag.content)){
				if(!strcasecmp(tag.equiv,"keywords")){
					strncpy(keywords,tag.content,sizeof(keywords));
					e=strtok(tag.content,SEPARATOR);
					while(e){
						if(index&&keywordweight)addword(e,keywordweight,1);
						e=strtok(NULL,SEPARATOR);
					}
				}else
				if(!strcasecmp(tag.equiv,"description")){
					strncpy(descript,tag.content,sizeof(descript));
					e=strtok(tag.content,SEPARATOR);
					while(e){
						if(index&&descweight)addword(e,descweight,1);
						e=strtok(NULL,SEPARATOR);
					}
				}else
				if(!strcasecmp(tag.equiv,"robots")){
					if(userobots){
					e=strtok(tag.content,SEPARATOR);
					while(e){
						if(!strcasecmp(e,"ALL")){
							follow=1;index=1;
						}else
						if(!strcasecmp(e,"NONE")){
							follow=0;index=0;
						}else
						if(!strcasecmp(e,"NOINDEX"))
							index=0;
						else
						if(!strcasecmp(e,"NOFOLLOW"))
							follow=0;
						else
						if(!strcasecmp(e,"INDEX"))
							index=1;
						else
						if(!strcasecmp(e,"FOLLOW"))
							follow=1;
						e=strtok(NULL,SEPARATOR);
					}}
					
				}
			}
			if((href)&&(hops<maxhops)&&(follow)&&(gfollow)){
				if(!parseurl(&newurl,href)){
					sprintf(str,"%s://%s%s%s",
					strcmp(newurl.schema,"")?newurl.schema:from.schema,
					strcmp(newurl.hostinfo,"")?newurl.hostinfo:from.hostinfo,
					strcmp(newurl.path,"")?newurl.path:from.path,
					newurl.filename);
					if(!STRNCMP(str,"ftp://")&&(strstr(str,";type=")))
						*(strstr(str,";type"))=0;

					if((!is_denied(str))&&((findrobots(str)<0)||(!userobots))){
						if(findserver(str)>=0)
							addurl(str,url_id,hops+1);
					}else{
						/*printf("denied %s\n",str);*/
					}
					freeurl(&newurl);
				}else{
					/*printf("not parsed %s\n",str);*/
				}
				free_it(href);
			}
			freetag(&tag);free(tmp);
			s=s+len; state=HTML_UNK;
			break;

		case HTML_TXT: /* text */
			e=strchr(s,'<');
			if(e) len=e-s;
			else  len=strlen(s);
			tmp=(char*)malloc(2*len+6);
			e=tmp;j=0;*e=0;
			for(i=0;i<len;i++){
				switch(s[i]){
				case ' ' : case '\t':
				case '\n': case '\r':
					j++; break;

				case '&': /* parse specials */
					if(!STRNCASECMP(s+i,"&nbsp;")){
						j++; i+=5; break;
					}
					if(!STRNCASECMP(s+i,"&nbsp")){
						j++; i+=4; break;
					}
					if(j){*e=' ';e++;j=0;}
					if(!STRNCASECMP(s+i,"&lt;")){
						*e='<';e++;*e=0; i+=3;
					}else
					if(!STRNCASECMP(s+i,"&gt;")){
						*e='>';e++;*e=0; i+=3;
					}else
					if(!STRNCASECMP(s+i,"&quot;")){
						*e='"';e++;*e=0; i+=5;
					}else
					if(!STRNCASECMP(s+i,"&lt")){
						*e='<';e++;*e=0; i+=2;
					}else
					if(!STRNCASECMP(s+i,"&gt")){
						*e='>';e++;*e=0; i+=2;
					}else
					if(!STRNCASECMP(s+i,"&quot")){
						*e='"';e++;*e=0; i+= 4;
					}else{
						*e='&';e++;*e=0;
					}
					break;
				default:
					if(j){*e=' ';e++;j=0;}
					*e=(s[i]=='\'')?'`':s[i];
					e++; *e=0;
					break;
				}
			}

			if((inbody&&!inscript)&&(strlen(text)<MAXTEXTSIZE)){
				if(*text)strcat(text," ");
				strncat(text,tmp,MAXTEXTSIZE);
			}

			if((intitle)&&(strlen(title)<MAXTITLESIZE)){
				if(*title)strcat(title," ");
				strncat(title,tmp,MAXTITLESIZE);
			}

			if(index){
				e=strtok(tmp,SEPARATOR);
				while(e){
					if(inbody&&bodyweight&&!inscript)
						addword(e,bodyweight,1);
					if(intitle&&titleweight)
						addword(e,titleweight,1);
					e=strtok(NULL,SEPARATOR);
				}
			}
			free(tmp);
			s=s+len; state=HTML_UNK;
			break;

		}} /* switch and while */
	}else{
		fprintf(stdout,"%s\n",content_type?content_type:"");
	}

	if((urlweight)&&(strcmp(from.filename,"robots.txt"))){
		s=strtok(cururl,SEPARATOR);
		while(s){
			addword(s,urlweight,1);
			s=strtok(NULL,SEPARATOR);
		}
	}
	storewords(url_id);
	freewords();

	/* Prepare text,keywords,description */
	s=text;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
	s=keywords;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
	s=descript;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
	mysql_escape_string(text_escaped,text,strlen(text));
	mysql_escape_string(keywords_escaped,keywords,strlen(keywords));
	mysql_escape_string(descript_escaped,descript,strlen(descript));

	/* Now save URL */
	sprintf(qbuf,"UPDATE url SET status='%d',last_index_time=NOW(),next_index_time=from_unixtime(unix_timestamp()+%d),tag=%d,text='%s',title='%s',content_type='%s',size=%d,indexed=1,keywords='%s',description='%s' WHERE rec_id='%d'",
		status,period,hint,text_escaped,title,content_type,
		realsize,keywords_escaped,descript_escaped,url_id);

	if(safe_mysql_query(&mysql,qbuf))abort_indexer();
	freeurl(&from);

} /* endwhile */

} /* end of program */
